class UnpackingVolumeConfiguration:
    UNPACKING_VOLUME_PATH: str = "_storage/unpacking"  # path to volume where unpacking will happen


class ProductionUnpackingVolumeConfiguration:
    UNPACKING_VOLUME_PATH: str = None
